/*---------------------------------------------------------------------------*
 |              PDFlib - A library for generating PDF on the fly             |
 +---------------------------------------------------------------------------+
 | Copyright (c) 1997-2007 Thomas Merz and PDFlib GmbH. All rights reserved. |
 +---------------------------------------------------------------------------+
 |                                                                           |
 |    This software is subject to the PDFlib license. It is NOT in the       |
 |    public domain. Extended versions and commercial licenses are           |
 |    available, please check http://www.pdflib.com.                         |
 |                                                                           |
 *---------------------------------------------------------------------------*/

/* $Id: pc_parfile.h,v 1.4.2.5 2007/03/08 19:50:43 kurt Exp $
 *
 * Parameter file parsing
 *
 */

#ifndef PC_PARFILE_H
#define PC_PARFILE_H

#include "pc_util.h"
#include "pc_file.h"
#include "pc_contain.h"

/* Parameter files
 * ---------------
 *
 * Those files have following structure:
 *
 * [name1]
 * item1: value1
 * item2: value2
 * ...
 * [name2]
 * item1: value1
 * item2: value2
 * ...
 *
 * Remarks:
 *
 * - "[name]" marks the begin of a so called "section".
 *   "name" is the name of the section.
 *   A name must be unique within a parameter file.
 *   A name can be empty: [].
 *   If a section is missing before the first item,
 *   a section with empty name will be assumed.
 *
 * - A section consists in a list of item names with corresponding
 *   item values (strings).
 *
 * - Item names and item values can be separated by arbitrary
 *   lot of following characters: ' ', '\t', '\v', '=', ':'.
 *   Therefore an item name cannot contain these characters.
 *
 * - A double quotation mark '"' at the begin or at the end of an
 *   item value will be skipped. A quotation mark must be used,
 *   if the item value begins or ends resp. with one or more
 *   of the above mentioned characters.
 *
 * - A backslash '\' at the end of an item value will be skipped.
 *   It marks the continuation of the value in the next line.
 *
 * - Empty lines and lines beginning with '%' will be ignored.
 *
 * - A logging will be obtained by class resource=1/2.
 *
 */

/*
 * Read a parameter file and save all sections in o_sects.
 * The file will be searched by resource SearchPath.
 * o_sects must be deleted by pdc_vtr_delete(o_sects);
 *
 * Return value: number of sections.
 *           -1: an error was occured
 *               which can be thrown by PDC_RETHROW():
 */
int pdc_parse_parfile(pdc_core *pdc, const char *filename, pdc_vtr **o_sects);


/* Return the index (beginning with 0) of section <name>.
 * -1: section not existing.
 */
int pdc_get_sectionnumber(pdc_vtr *sects, const char *name);


/* Return the number of items of section <is>.
 * -1: section not existing.
 */
int pdc_get_numitems(pdc_vtr *sects, int is);


/* Return the name of section <is>.
 * NULL: section not existing.
 */
const char *pdc_get_sectionname(pdc_vtr *sects, int is);


/* Return the name of item with index <it> (beginning with 0) of section <is>.
 * NULL: section or item not existing.
 */
const char *pdc_get_itemname(pdc_vtr *sects, int is, int it);


/* Return the value of item <name> (name != NULL) or <it> of section <is>.
 * NULL: section or item not existing.
 */
const char *pdc_get_itemvalue(pdc_vtr *sects, int is, const char *name, int it);


/* Insert the item with index <it>, with name <item> and with the value <value>
 * into section <is>.
 * If <it> == -1, a new item will be created and its index will be returned.
 * -1: section not existing.
 */
int
pdc_insert_item(pdc_core *pdc, pdc_vtr *sects, int is, int it,
        const char *item, const char *value);


/* Delete the item with index <it> in section <is>.
 * The new number of items will be returned.
 * -1: section not existing.
 */
int
pdc_delete_item(pdc_core *pdc, pdc_vtr *sects, int is, int it);

#endif /* PC_PARFILE_H */
