/*---------------------------------------------------------------------------*
 |              PDFlib - A library for generating PDF on the fly             |
 +---------------------------------------------------------------------------+
 | Copyright (c) 1997-2007 Thomas Merz and PDFlib GmbH. All rights reserved. |
 +---------------------------------------------------------------------------+
 |                                                                           |
 |    This software is subject to the PDFlib license. It is NOT in the       |
 |    public domain. Extended versions and commercial licenses are           |
 |    available, please check http://www.pdflib.com.                         |
 |                                                                           |
 *---------------------------------------------------------------------------*/

/* $Id: fonttabs.c,v 1.1.2.12 2007/03/06 11:34:18 kurt Exp $
 *
 * Font tables to show font and encoding content
 */

#include <math.h>
#include "fonttabs.h"
#include "p_intern.h"

static const char *embedtext[] = {"will not embed", "will embed"};
static const char *nativetext[] = {"nokeepnative", "keepnative"};

#define MAXNAMELEN 128

/*
 * Get font and encoding name
 */
static void
get_font_names(PDF *p, int testfont, char *userfontname, char *fontname,
               char *userencname, char *encname)
{
    double info;

    /* fontname */
    info = PDF_info_font(p, testfont, "fontname", "api");
    strcpy(userfontname, PDF_get_parameter(p, "string", (int) info));
    info = PDF_info_font(p, testfont, "fontname", "full");
    strcpy(fontname, PDF_get_parameter(p, "string", (int) info));

    /* encoding */
    info = PDF_info_font(p, testfont, "encoding", "api");
    strcpy(userencname, PDF_get_parameter(p, "string", (int) info));
    info = PDF_info_font(p, testfont, "encoding", "actual");
    strcpy(encname, PDF_get_parameter(p, "string", (int) info));
}

/*
 * Show the properties of a font
 */
static void
show_font_property(
    PDF *p,
    int testfont,
    const char *keyword,
    const char *optlist,
    int isunicode,
    int isstring)
{
    const double LEFT = 20.0;
    const double TABX = 215;
    char buf[256];
    double y, prop;

    strcpy(buf, keyword);
    if (strlen(optlist))
    {
        strcat(buf, " (");
        strcat(buf, optlist);
        strcat(buf, ")");
    }
    strcat(buf, ":  ");
    PDF_continue_text(p, buf);
    buf[0] = 0;

    prop = PDF_info_font(p, testfont, keyword, optlist);
    if (isstring)
    {
        if (prop > -1)
            strcat(buf, PDF_get_parameter(p, "string", (int) prop));
    }
    else if (isunicode)
    {
        if (prop > -1)
            sprintf(buf, "U+%04X", (int) prop);
    }
    else
    {
        sprintf(buf, "%g", prop);
    }

    y = PDF_get_value(p, "texty", 0);
    PDF_show_xy(p, buf, TABX, y);
    PDF_set_text_pos(p, LEFT, y);
}

void
pdf_show_font_properties(PDF *p, int testfont)
{
    const double PWIDTH = a4_width;
    const double PHEIGHT = a4_height;

    const double TOPY = PHEIGHT - 40.0;
    const double TABY = TOPY - 20.0;
    const double LEFT = 20.0;
    const double BIGFRAMEFONTSIZE = 16.0;
    const double FRAMEFONTSIZE = 14.0;
    char buf[256];
    char userfontname[MAXNAMELEN], fontname[MAXNAMELEN];
    char userencname[MAXNAMELEN], encname[MAXNAMELEN];
    int framefont;

    /* get fontname and encoding */
    get_font_names(p, testfont, userfontname, fontname, userencname, encname);

    framefont = PDF_load_font(p, PDF_FRAMEFONTNAME, 0, PDF_FRAMEFONTENC,
                              PDF_FRAMEFONTOPT);

    PDF_begin_page_ext(p, PWIDTH, PHEIGHT, "");

    /* bookmark */
    sprintf(buf, "%s Property List", userfontname);
    PDF_create_bookmark(p, buf, 0, "");

    /* headline */
    PDF_setfont(p, framefont, BIGFRAMEFONTSIZE);
    sprintf(buf, "%s", fontname);
    PDF_show_xy(p, buf, LEFT, TOPY);

    PDF_setfont(p, framefont, FRAMEFONTSIZE);
    PDF_set_value(p, "leading", 1.18 * FRAMEFONTSIZE);
    PDF_show_xy(p, "", LEFT, TABY);

    show_font_property(p, testfont, "fontname", "api", 0, 1);
    show_font_property(p, testfont, "fontname", "acrobat", 0, 1);
    show_font_property(p, testfont, "fontname", "full", 0, 1);

    show_font_property(p, testfont, "fontfile", "", 0, 1);
    show_font_property(p, testfont, "metricsfile", "", 0, 1);
    show_font_property(p, testfont, "hostfont", "", 0, 0);

    show_font_property(p, testfont, "encoding", "api", 0, 1);
    show_font_property(p, testfont, "encoding", "actual", 0, 1);
    show_font_property(p, testfont, "supplement", "", 0, 0);

    show_font_property(p, testfont, "symbolfont", "", 0, 0);
    show_font_property(p, testfont, "cidfont", "", 0, 0);
    show_font_property(p, testfont, "unicodefont", "", 0, 0);
    show_font_property(p, testfont, "standardfont", "", 0, 0);

    show_font_property(p, testfont, "willsubset", "", 0, 0);
    show_font_property(p, testfont, "willembed", "", 0, 0);
    show_font_property(p, testfont, "keepnative", "", 0, 0);

    show_font_property(p, testfont, "vertical", "", 0, 0);
    show_font_property(p, testfont, "fontstyle", "", 0, 1);
    show_font_property(p, testfont, "fontstyle", "faked", 0, 0);
    show_font_property(p, testfont, "italicangle", "", 0, 0);
    show_font_property(p, testfont, "monospace", "", 0, 0);
    show_font_property(p, testfont, "weight", "", 0, 0);

    show_font_property(p, testfont, "numglyphs", "", 0, 0);
    show_font_property(p, testfont, "numcids", "", 0, 0);
    show_font_property(p, testfont, "maxcode", "", 0, 0);
    show_font_property(p, testfont, "replacementchar", "", 1, 0);
    show_font_property(p, testfont, "kerningpairs", "", 0, 0);
    show_font_property(p, testfont, "unmappedglyphs", "", 0, 0);

    show_font_property(p, testfont, "ascender", "fontsize=1000", 0, 0);
    show_font_property(p, testfont, "ascender", "fontsize=10", 0, 0);
    show_font_property(p, testfont, "ascender", "faked", 0, 0);
    show_font_property(p, testfont, "capheight", "fontsize=1000", 0, 0);
    show_font_property(p, testfont, "capheight", "fontsize=10", 0, 0);
    show_font_property(p, testfont, "capheight", "faked", 0, 0);
    show_font_property(p, testfont, "xheight", "fontsize=1000", 0, 0);
    show_font_property(p, testfont, "xheight", "fontsize=10", 0, 0);
    show_font_property(p, testfont, "xheight", "faked", 0, 0);
    show_font_property(p, testfont, "descender", "fontsize=1000", 0, 0);
    show_font_property(p, testfont, "descender", "fontsize=10", 0, 0);
    show_font_property(p, testfont, "descender", "faked", 0, 0);
    show_font_property(p, testfont, "linegap", "fontsize=1000", 0, 0);
    show_font_property(p, testfont, "linegap", "fontsize=10", 0, 0);
    show_font_property(p, testfont, "linegap", "faked", 0, 0);

    PDF_end_page(p);
}


/*
 * Show for a font an encoding
 */
void
pdf_show_font_encoding(PDF *p, int testfont)
{
    pdf_show_font_encoding_ext(p, testfont, 0, NULL);
}

void
pdf_show_font_encoding_ext(PDF *p, int testfont,
                           int unistart, const char *comment)
{
    const double PWIDTH = a4_width;
    const double PHEIGHT = a4_height;

    const double TOPY = PHEIGHT - 40.0;
    const double TABY = TOPY - 50.0;
    const double LEFT = 35.0;
    const double FONTSIZE = 24.0;
    const double XINCR = 32;
    const double YINCR = 36;
    const double FRAMEFONTSIZE = 12.0;
    const double BIGFRAMEFONTSIZE = 16.0;
    const double SMALLFRAMEFONTSIZE = 10.0;
    const double BASE = 100.0;
    const double MFONTSIZE = 48.0;
    const double RIGHT = a4_width - LEFT;

    const pdc_byte TESTSTRING[] = "\x54\x48\xC4\x66\x64\x6A\x70\x6F\x78";
                                  /* THfdjpox */

    const pdc_ushort UNITESTSTR[] =
    {
        0x0054, 0x0048, 0x00C4, 0x0066, 0x0064, 0x006A, 0x0070, 0x006F, 0x0078
    };

    const pdc_ushort GLYTESTSTR[] = {5,6,7,8,9,10,11,12,13};

    const double LEGENDCOLCOMP = 1.0;
    const double DIMCOLCOMP = 0.6;
    const double GRAYCOLCOMP = 0.8;

    const char *glyphname = "space";
    const char *chartype = "";
    const char *codepage = NULL;
    const char *surrpair = NULL;
    const char *legoptlist = "position {100 0}";
    char buf[256], buf1[256], optlist[256], fontstyle[32], teststring[64];
    char userfontname[MAXNAMELEN], fontname[MAXNAMELEN];
    char userencname[MAXNAMELEN], encname[MAXNAMELEN];
    pdc_ushort uv;
    pdc_bool vertical = 0, isunicode = 0, isglyphid = 0;
    pdc_bool is8bit = 0, isbuiltin = 0, isempty = 0, diffenc = 0;
    pdc_scalar ascender, descender, capheight, xheight;
    pdc_scalar x, y, ty, xincr, yincr, yincr2, fontsize;
    pdc_scalar width, height, widthrect, red = 0, green = 0, blue = 0;
    int numcodes, page, numpages, code, numglyphs = 0, monospace = 0;
    int gid = 1, usv = 1, metricpage = 0, numrows = 0, row, col, i, n;
    int len, nbytes = 1, bookmark = 0, willembed = 0, framefont, pageno;
    double info;

    /* get fontname and encoding */
    get_font_names(p, testfont, userfontname, fontname, userencname, encname);
    diffenc = strcmp(userencname, encname);

    /* get fontstyle */
    info = PDF_info_font(p, testfont, "fontstyle", "");
    strcpy(fontstyle, PDF_get_parameter(p, "string", (int) info));

    isbuiltin = !strcmp(encname, "builtin");
    if (!isbuiltin)
        isunicode = !strcmp(userencname, "unicode");
    is8bit = isunicode && strcmp(encname, "unicode");
    isglyphid = !strcmp(encname, "glyphid");

    vertical = (pdc_bool) PDF_info_font(p, testfont, "vertical", "");
    monospace = (int) PDF_info_font(p, testfont, "monospace", "");
    willembed = (int) PDF_info_font(p, testfont, "willembed", "");

    framefont = PDF_load_font(p, PDF_FRAMEFONTNAME, 0, PDF_FRAMEFONTENC,
                              PDF_FRAMEFONTOPT);

    PDF_set_parameter(p, "charref", "false");
    PDF_set_parameter(p, "escapesequence", "false");
    PDF_set_parameter(p, "glyphcheck", "none");
    PDF_set_parameter(p, "textformat", "auto");

    if (is8bit)
        numcodes = 0x10000;
    else
        numcodes = (int) PDF_info_font(p, testfont, "maxcode", "") + 1;
    numpages = numcodes / 256;
    if (256 * numpages < numcodes)
        numpages++;

    if (isunicode || isglyphid)
        nbytes = 2;

    height = FONTSIZE * PDF_get_value(p, "capheight", (double) testfont);
    fontsize = FONTSIZE;
    xincr = XINCR;
    yincr = YINCR;
    yincr2 = 1.2 * SMALLFRAMEFONTSIZE;

    len = nbytes * strlen(TESTSTRING);
    if (isunicode)
    {
        memcpy(teststring, UNITESTSTR, len);
        chartype = "unicode";
    }
    else if (isglyphid)
    {
        memset(teststring, 0, len);
        chartype = "id";
        metricpage = numpages - 1;
        isempty = 1;
    }
    else
    {
        strcpy(teststring, TESTSTRING);
    }

    /* loop over all pages */
    for (page = 0; page < numpages; page++)
    {
        if (page)
        {
            pageno = page << 8;
            sprintf(buf1, "x%04X", pageno);
            if (!isglyphid)
            {
                if (pageno == 0x0100)
                    codepage = "Latin";
                else if (pageno == 0x0300)
                    codepage = "Latin/Greek/Coptic";
                else if (pageno == 0x0400)
                    codepage = "Cyrillic";
                else if (pageno == 0x0500)
                    codepage = "Armenian/Hebrew";
                else if (pageno == 0x0600)
                    codepage = "Arabic";
                else if (pageno == 0x0700)
                    codepage = "Syrian/Thaana";
                else if (pageno == 0x0800)
                    codepage = NULL;
                else if (pageno == 0x0900)
                    codepage = "Devanagari/Bengali";
                else if (pageno == 0x0A00)
                    codepage = "Gurmukhi/Gurajati";
                else if (pageno == 0x0B00)
                    codepage = "Oriya/Tamil";
                else if (pageno == 0x0C00)
                    codepage = "Telugu/Kannada";
                else if (pageno == 0x0D00)
                    codepage = "Malayalam/Sinhala";
                else if (pageno == 0x0E00)
                    codepage = "Thai/Lao";
                else if (pageno == 0x0F00)
                    codepage = "Tibetan";
                else if (pageno == 0x1000)
                    codepage = "Myanmar/Georgian";
                else if (pageno == 0x1100)
                    codepage = "Hangul Jamo";
                else if (pageno == 0x1200)
                    codepage = "Ethiopic";
                else if (pageno == 0x1300)
                    codepage = "Ethiopic/Cherokee";
                else if (pageno == 0x1400)
                    codepage = "Canadian Aboriginal";
                else if (pageno == 0x1600)
                    codepage = "Canadian Aboriginal/Ogham/Runic";
                else if (pageno == 0x1700)
                    codepage = "Khmer";
                else if (pageno == 0x1800)
                    codepage = "Mongolian";
                else if (pageno == 0x1E00)
                    codepage = "Extended Latin";
                else if (pageno == 0x1F00)
                    codepage = "Extended Greek";
                else if (pageno == 0x2000)
                    codepage = "Symbols";
                else if (pageno == 0x2800)
                    codepage = "Symbols/CJK Misc";
                else if (pageno == 0x2900)
                    codepage = "CJK Misc";
                else if (pageno == 0x3000)
                    codepage = "CJK Misc/Hiragana/Katakana";
                else if (pageno == 0x3100)
                    codepage = "Bopomofo/Hangul/Kanbun";
                else if (pageno == 0x3200)
                    codepage = "Enclosed CJK Letters";
                else if (pageno == 0x3300)
                    codepage = "CJK Compatibility";
                else if (pageno == 0x3400)
                    codepage = "CJK Unified Ideographs";
                else if (pageno == 0xA000)
                    codepage = "Yi Syllables";
                else if (pageno == 0xAC00)
                    codepage = "Hangul Syllables";
                else if (pageno >= 0xD800 && pageno < 0xE000) /* Surrogates */
                    continue;
                else if (pageno == 0xE000)
                    codepage = "Private Use Area";
                else if (pageno == 0xF900)
                    codepage = "CJK Ideographs";
                else if (pageno == 0xFB00)
                    codepage = "Presentation Forms";
                else if (pageno == 0xFF00)
                    codepage = "Halfwidth and Fullwidths Forms";
                else if (pageno == 0x10000)
                    codepage = "Archaeological characters";
                else if (pageno == 0x1D000)
                    codepage = "Musical Symbols";
                else if (pageno == 0x1D300)
                    codepage = "Tai Xuan Jing Symbols";
                else if (pageno == 0x1D400)
                    codepage = "Mathematical Aplphanumeric Symbols";
                else if (pageno == 0x20000)
                    codepage = "CJK Unified Ideographs Extension B";

                /* skip empty page for unicode encoding */
                if (isunicode)
                {
                    for (i = 0; i < 256; i++)
                    {
                        code = page * 256 + i;
                        sprintf(buf, "unicode=0x%05X", code);

                        if (is8bit)
                        {
                            gid = (int) PDF_info_font(p, testfont, "code", buf);
                            if (gid <= 0)
                                continue;
                            sprintf(buf, "code=%d", gid);
                        }
                        gid = (int) PDF_info_font(p, testfont, "glyphid", buf);
                        if (gid > 0)
                            break;
                    }
                    if (i == 256)
                        continue;
                }
                if (codepage)
                    sprintf(buf, "%s  (%s)", buf1, codepage);
                else
                    sprintf(buf, "%s", buf1);
            }
            else
            {
                sprintf(buf, "%s  (%d)", buf1, pageno);
            }
        }

        PDF_begin_page_ext(p, PWIDTH, PHEIGHT, "");
        PDF_setlinewidth(p,  0.2);

        /* real encoding name for testfont */
        PDF_setfont(p, testfont, fontsize);

        /* bookmark */
        if (!page)
        {
            strcpy(buf1, "x0000");
            strcpy(buf, userfontname);
            if (strcmp(fontstyle, "Normal"))
            {
                strcat(buf, ",");
                strcat(buf, fontstyle);
            }
            strcat(buf, " (");
            strcat(buf, userencname);
            strcat(buf, ") ");
            if (monospace)
                sprintf(&buf[strlen(buf)], "(monospace %d) ", monospace);
            if (comment != NULL && strlen(comment))
            {
                strcat(buf, " (");
                strcat(buf, comment);
                strcat(buf, ") ");
            }
            if (numpages > 1)
                strcat(buf, buf1);
            bookmark = PDF_create_bookmark(p, buf, 0, "");
        }
        else
        {
            sprintf(optlist, "parent %d  open", bookmark);
            PDF_create_bookmark(p, buf, 0, optlist);
        }

        /* headline */
        PDF_setfont(p, framefont, BIGFRAMEFONTSIZE);
        y = TOPY;
        sprintf(buf, "%s", fontname);
        if (strcmp(fontstyle, "Normal"))
        {
            strcat(buf, ",");
            strcat(buf, fontstyle);
        }
        PDF_show_xy(p, buf, LEFT, y);

        PDF_setfont(p, framefont, FRAMEFONTSIZE);

        /* encoding, embedding */
        if (!diffenc)
            sprintf(buf, "  (%s,  %s)", encname, embedtext[willembed]);
        else
            sprintf(buf, "  (%s => %s,  %s)",
                    userencname, encname, embedtext[willembed]);
        if (vertical)
            strcat(buf, ", vertical");
        PDF_show(p, buf);

        /* glyphs */
        if (numpages > 1)
        {
            sprintf(buf, "Glyphs from %s %s +", chartype, buf1);
            PDF_continue_text(p, NULL);
            PDF_continue_text(p, buf);
        }

        y = TABY;

        if (page < unistart)
        {
            PDF_show_xy(p, "skipped", LEFT, y);
            numrows = 0;
        }
        else
        {
            /* column numbering */
            for (col = 0; col < 16; col++)
            {
                sprintf(buf, "x%X", col);
                PDF_show_xy(p, buf, LEFT + (col + 1) * xincr, y);
            }

            /* number of rows */
            numrows = 16;
            if (numcodes < (page + 1) * 256)
            {
                n = numcodes - page * 256;
                numrows = n / 16;
                if (16 * numrows < n) numrows++;
            }
        }

        /* row loop */
        for (row = 0; row < numrows; row++)
        {
            y -= yincr;
            x = LEFT;

            /* row numbering */
            PDF_setfont(p, framefont, FRAMEFONTSIZE);
            sprintf(buf, "%Xx", row);
            PDF_show_xy(p, buf, LEFT, y);

            /* set testfont */
            PDF_setfont(p, testfont, fontsize);

            /* show glyphs per row */
            for (col = 0; col < 16; col++)
            {
                /* character */
                code = page * 256 + (16 * row + col);
                if (code == numcodes)
                    break;

                /* glyphid */
                if (isglyphid)
                {
                    gid = code;

                    sprintf(buf, "glyphid=%d", gid);
                    usv = (int) PDF_info_font(p, testfont, "unicode", buf);

                    for (i = 0; i < len/2; i++)
                    {
                        if ((int) UNITESTSTR[i] == usv)
                        {
                            uv = (pdc_ushort) code;
                            memcpy(&teststring[2*i], &uv, 2);
                            isempty = 0;
                            break;
                        }
                    }
                    numglyphs++;
                }

                /* unicode */
                else if (isunicode)
                {
                    usv = code;

                    gid = 1;
                    sprintf(buf, "unicode=0x%05X", usv);
                    if (is8bit)
                    {
                        gid = (int) PDF_info_font(p, testfont, "code", buf);
                        if (gid > 0)
                            sprintf(buf, "code=%d", gid);
                    }
                    if (gid > 0)
                        gid = (int) PDF_info_font(p, testfont, "glyphid", buf);
                }

                /* builtin */
                else if (isbuiltin)
                {
                    sprintf(buf, "code=%d", code);
                    gid = (int) PDF_info_font(p, testfont, "glyphid", buf);
                }

                /* encoding */
                else
                {
                    sprintf(buf, "code=%d", code);
                    gid = (int) PDF_info_font(p, testfont, "glyphid", buf);

                    usv = (int) PDF_info_font(p, testfont, "unicode", buf);
                }

                /* glyphname */
                if (!isbuiltin)
                {
                    glyphname = NULL;
                    if (usv >= 0)
                    {
                        sprintf(buf, "unicode=0x%05X", usv);
                        info = PDF_info_font(p, testfont, "glyphname", buf);
                        if (info > -1)
                            glyphname = PDF_get_parameter(p, "string", (int) info);
                    }
                }

                /* show character box and character */
                x += xincr;

                /* glyph available */
                if (gid >= 0)
                {
                    /* text string */
                    if (nbytes == 1)
                    {
                        buf[0] = (char) code;
                        buf[1] = 0;
                    }
                    else if (code <= 0xFFFF)
                    {
                        nbytes = 2;
                        uv = (pdc_short) code;
                        memcpy(buf, &uv, nbytes);
                    }
                    else
                    {
                        surrpair =
                            PDF_utf32_to_utf16(p, (char *) &code, 4, "", &nbytes);
                        memcpy(buf, surrpair, nbytes);
                    }

                    /* glyph width */
                    width = PDF_info_textline(p, buf, nbytes, "width", "");

                    if (usv < 0)
                    {
                        red = LEGENDCOLCOMP;
                        green = blue = DIMCOLCOMP;
                    }
                    else if (glyphname == NULL)
                    {
                        red = green = DIMCOLCOMP;
                        blue = LEGENDCOLCOMP;
                    }
                    else
                    {
                        red = green = blue = GRAYCOLCOMP;
                    }

                    /* test rectangle */
                    PDF_setcolor(p, "fillstroke", "rgb", red, green, blue, 0);
                    widthrect = (width != 0) ? width : height;
                    PDF_rect(p, x, y, widthrect, height);
                    if (width != 0)
                        PDF_fill(p);
                    else
                        PDF_stroke(p);

                    /* glyph */
                    PDF_setcolor(p, "fillstroke", "gray", 0, 0, 0, 0);
                    PDF_fit_textline(p, buf, nbytes, x, y, "");
                }
                else
                {
                    PDF_setcolor(p, "fillstroke", "gray", GRAYCOLCOMP, 0, 0, 0);
                    PDF_rect(p, x, y, height, height);
                    PDF_stroke(p);
                    PDF_setcolor(p, "fillstroke", "gray", 0, 0, 0, 0);
                }
            }
        }

        PDF_setfont(p, framefont, SMALLFRAMEFONTSIZE);

        PDF_setcolor(p, "fill", "rgb",
                     GRAYCOLCOMP, GRAYCOLCOMP, GRAYCOLCOMP, 0);
        ty = BASE - 3.0/4.0 * MFONTSIZE;
        PDF_fit_textline(p, "Glyph completely defined", 0, RIGHT, ty,
                         legoptlist);

        PDF_setcolor(p, "fill", "rgb",
                     LEGENDCOLCOMP, DIMCOLCOMP, DIMCOLCOMP, 0);
        ty -= yincr2;
        PDF_fit_textline(p, "Unicode mapping not found", 0, RIGHT, ty,
                         legoptlist);

        PDF_setcolor(p, "fill", "rgb",
                     DIMCOLCOMP, DIMCOLCOMP, LEGENDCOLCOMP, 0);
        ty -= yincr2;
        PDF_fit_textline(p, "Registered glyph name not found", 0, RIGHT, ty,
                         legoptlist);

        PDF_setcolor(p, "fill", "rgb", 0, 0, 0, 0);

        /* last page */
        if (page == metricpage)
        {
            if (isglyphid)
            {
                PDF_setfont(p, framefont, BIGFRAMEFONTSIZE);
                numcodes = (int) PDF_info_font(p, testfont, "numglyphs","");
                sprintf(buf, "Number of glyphs: %d (found: %d)",
                        numcodes, numglyphs);
                PDF_show_xy(p, buf, LEFT, y - 2 * BIGFRAMEFONTSIZE);
                if (isempty)
                    memcpy(teststring, GLYTESTSTR, len);
            }

            sprintf(buf, "font=%d fontsize=%g glyphcheck=replace",
                    testfont, MFONTSIZE);
            width = PDF_info_textline(p, teststring, len, "width", buf);

            ascender = PDF_get_value(p, "ascender", (double) testfont);
            descender = PDF_get_value(p, "descender", (double) testfont);
            capheight = PDF_get_value(p, "capheight", (double) testfont);
            xheight = PDF_get_value(p, "xheight", (double) testfont);

            /* capheight */
            PDF_setcolor(p, "fill", "gray", 0.8, 0, 0, 0);
            PDF_rect(p, LEFT, BASE, width, MFONTSIZE * capheight);
            PDF_fill(p);

            /* ascender */
            PDF_setdash(p, 5, 5);
            PDF_setcolor(p, "stroke", "gray", 0.3, 0, 0, 0);
            PDF_moveto(p, LEFT, BASE + MFONTSIZE * ascender);
            PDF_lineto(p, LEFT + width, BASE + MFONTSIZE * ascender);
            PDF_stroke(p);

            /* descender */
            PDF_setdash(p, 3, 3);
            PDF_setcolor(p, "stroke", "gray", 0.3, 0, 0, 0);
            PDF_moveto(p, LEFT, BASE + MFONTSIZE * descender);
            PDF_lineto(p, LEFT + width, BASE + MFONTSIZE * descender);
            PDF_stroke(p);

            /* xheight */
            PDF_setdash(p, 2, 2);
            PDF_setcolor(p, "stroke", "gray", 0.3, 0, 0, 0);
            PDF_moveto(p, LEFT, BASE + MFONTSIZE * xheight);
            PDF_lineto(p, LEFT + width, BASE + MFONTSIZE * xheight);
            PDF_stroke(p);

            /* glyphs */
            PDF_setcolor(p, "fill", "gray", 0, 0, 0, 0);

            /* test text */
            PDF_fit_textline(p, teststring, len, LEFT, BASE, buf);

            /* font metrics */
            PDF_setfont(p, framefont, SMALLFRAMEFONTSIZE);

            PDF_set_text_pos(p, LEFT, BASE - 3.0/4.0 * MFONTSIZE);
            sprintf(buf, "ascender = %d",
                    (int) floor(1000 * ascender + 0.5));
            if (!strcmp(PDF_get_parameter(p, "ascenderfaked",
                        (double) testfont), "true"))
                strcat(buf, "   (faked)");
            PDF_show(p, buf);

            PDF_set_value(p, "leading", (double) SMALLFRAMEFONTSIZE);
            sprintf(buf, "capheight = %d",
                    (int) floor(1000 * capheight + 0.5));
            if (!strcmp(PDF_get_parameter(p, "capheightfaked",
                        (double) testfont), "true"))
                strcat(buf, "   (faked)");
            PDF_continue_text(p, buf);

            sprintf(buf, "xheight = %d",
                    (int) floor(1000 * xheight + 0.5));
            if (!strcmp(PDF_get_parameter(p, "xheightfaked",
                        (double) testfont), "true"))
                strcat(buf, "   (faked)");
            PDF_continue_text(p, buf);

            sprintf(buf, "descender = %d",
                    (int) floor(1000 * descender + 0.5));
            if (!strcmp(PDF_get_parameter(p, "descenderfaked",
                        (double) testfont), "true"))
                strcat(buf, "   (faked)");
            PDF_continue_text(p, buf);
        }

        PDF_end_page(p);
    }
}

